/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Simple Timer + Clocks.
 *
 * The Initial Developer of the Original Code is
 *   George Bradt.
 *
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   George Bradt.
 *   Tsprajna.
 *
 * Credits:
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var SimpleClocksNotify = {
    // For reference, not used.
    notification: {completed: false,
                   recurring: false,
                   timeDisplay: "",
                   description: "",
                   time24hr: "",
                   timeMilli: 0,
                   timeAdjustedForDst: false,
                   currDay: -1,
                   day: -1,
                   url: "",
                   location: "",
                   sortValue: 0
                  },
    notifyTable: [],
    notifySep: "\x1D", // used to separate nots. in pref.
    locationsTable: [],

    // Called when loading Notify Me At dialog.

    onLoadNotify: function() {
        var strbundle = document.getElementById("simclk-strings");

        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                    getService(Components.interfaces.fuelIApplication);

        if ( Application.storage.has("notifyTable") ) {
            this.notifyTable = Application.storage.get("notifyTable", "ERR");

            if ( this.notifyTable === "ERR" ) {
                alert(strbundle.getString("alert.error.loading.notifications"));
                return;
            }
        }

        if ( Application.storage.has("locationsTable") ) {
            this.locationsTable = Application.storage.get("locationsTable", "ERR");

            if ( this.locationsTable === "ERR" ) {
                this.locationsTable = [];
                alert(strbundle.getString("alert.error.loading.locations"));
            }
        }

        this.populateLocaleMenulist();

        var treeChildren = document.getElementById("simclk-treeItems");

        // Clear the old tree data.
        this.onDeleteAllNotify();

        for ( var i in this.notifyTable ) {
            // Bypass completed notifications.
            if ( !this.notifyTable[i].completed ) {
                this.addTreeRow(treeChildren, this.notifyTable[i]);
            }
        }
    },

    // Called when user clicks Bookmark.

    onBookmarkNotify: function() {
        var strbundle = document.getElementById("simclk-strings");
        var url = window.opener.gBrowser.currentURI.spec;

        if ( url === "about:blank" ) {
            alert(strbundle.getString("alert.warning.no.url"));
        }
        else {
            document.getElementById("simclk-tboxNotifyUrl").value = url;
        }
    },

    // Called when user clicks Recurring checkbox.
    // Enable/disable day menulist.

    onClickRecurringNotify: function() {
        document.getElementById("simclk-mlistDow").disabled =
            !document.getElementById("simclk-cboxNotifyRecurring").checked;
    },

    // Called when user clicks on Notifications tree.

    onClickNotTree: function(event) {
        var strbundle = document.getElementById("simclk-strings");
        var locale, mlistLocale, recurring, dow, mlistDow;
        var tree = document.getElementById("simclk-treeNotifications");

        // See if tree disabled (ie. edit already in progress).
        if ( tree.disabled ) {
            return;
        }

        // Get the row, col and child element at the point of mouse click
        var tbo = tree.treeBoxObject;
        var row = { }, col = { }, child = { };
        tbo.getCellAt(event.clientX, event.clientY, row, col, child);

        if ( col.value ) {
            if ( col.value.id !== "simclk-colRecur" &&
                event.detail === 2 ) {
                // No monkey business.
                tree.disabled = true;

                // User double-clicked on a notification to initiate an edit.
                // First disable the delete and reset buttons.
                document.getElementById("simclk-btnDel").disabled = true;
                document.getElementById("simclk-btnDelAll").disabled = true;
                document.getElementById("simclk-btnReset").disabled = true;

                // Change label and tooltips on Add and Add/OK buttons.
                document.getElementById("simclk-btnAddOk").label =
                        strbundle.getString("button.label.replace/OK");
                document.getElementById("simclk-btnAddOk").tooltipText =
                        strbundle.getString("button.tooltip.replace/OK");
                document.getElementById("simclk-btnAdd").label =
                        strbundle.getString("button.label.replace");
                document.getElementById("simclk-btnAdd").tooltipText =
                        strbundle.getString("button.tooltip.replace");

                // Move the notification data back to the upper half.
                document.getElementById("simclk-tboxNotifyDesc").value =
                        tree.view.getCellText(row.value,
                        tree.columns.getNamedColumn("simclk-colDescription"));

                document.getElementById("simclk-tboxNotifyUrl").value =
                        tree.view.getCellText(row.value,
                        tree.columns.getNamedColumn("simclk-colUrl"));

                document.getElementById("simclk-tspinNotifyTime").value =
                        tree.view.getCellValue(row.value,
                        tree.columns.getNamedColumn("simclk-colTime"));

                mlistLocale = document.getElementById("simclk-mlistLocale");
                locale = tree.view.getCellText(row.value,
                        tree.columns.getNamedColumn("simclk-colLocale"));
                var localeFound = false;

                for ( var i = 0; i < mlistLocale.itemCount; i++ ) {
                    if ( mlistLocale.getItemAtIndex(i).label === locale ) {
                        mlistLocale.selectedIndex = i;
                        localeFound = true;
                        break;
                    }
                }

                if ( !localeFound ) {
                    alert(strbundle.getString("alert.warning.no.locale"));
                    mlistLocale.selectedIndex = 0;
                }

                mlistDow = document.getElementById("simclk-mlistDow");
                recurring = tree.view.getCellValue(row.value,
                        tree.columns.getNamedColumn("simclk-colRecur"));

                if ( recurring === "true" ) {
                    document.getElementById("simclk-cboxNotifyRecurring").checked = true;
                    dow = tree.view.getCellValue(row.value,
                            tree.columns.getNamedColumn("simclk-colDow"));
                    var dowFound = false;
                    var itemCount = mlistDow.itemCount;

                    for ( var j = 0; j < itemCount; j++ ) {
                         if ( mlistDow.getItemAtIndex(j).value === dow ) {
                            mlistDow.selectedIndex = j;
                            dowFound = true;
                            break;
                        }
                    }

                    if ( !dowFound ) {
                        alert(strbundle.getString("alert.error.invalid.dow"));
                        mlistDow.selectedIndex = 0;
                    }
                    else {
                        mlistDow.disabled = false;
                    }
                }
                else {
                    document.getElementById("simclk-cboxNotifyRecurring").checked = false;
                    mlistDow.disabled = true;
                }

                // Stash the row in a custom attribute.
                document.getElementById("simclk-tboxNotifyDesc").
                        setAttribute("updaterow", row.value);

                return;
            }

            if ( col.value.id === "simclk-colRecur" ) {
                // Toggling Recurring also changes Day.
                // Post-click value.
                if ( tree.view.getCellValue(row.value, col.value) === "true" ) {
                    tree.view.setCellText(row.value,
                            tree.columns.getNamedColumn("simclk-colDow"),
                            strbundle.getString("display.dow.daily"));
                    tree.view.setCellValue(row.value,
                            tree.columns.getNamedColumn("simclk-colDow"), "7");
                }
                else {
                    tree.view.setCellText(row.value,
                            tree.columns.getNamedColumn("simclk-colDow"), "");
                    tree.view.setCellValue(row.value,
                            tree.columns.getNamedColumn("simclk-colDow"), "-1");
                }

                return;
            }

            if ( col.value.id === "simclk-colUrl" &&
                 event.button === 2 ) {
                // User right-clicked on URL.
                var url = tree.view.getCellText(row.value, col.value);

                if ( url ) {
                    this.openLink(url);
                }
            }
        }
    },

    // Called when Add/OK button is clicked.

    onAddOkNotify: function() {
        if ( this.onAddNotify() ) {
            this.onOKNotify();
            window.close();
        }
    },

    // Called when Add button is clicked.

    onAddNotify: function() {
        // The timepicker element always stores its value as a 24 hr clock.
        // For display, convert value to locale format, and remove seconds.
        var strbundle = document.getElementById("simclk-strings");
        var time24hr = document.getElementById("simclk-tspinNotifyTime").value;
        var timeDisplay = new Date("December 23, 2007 " + time24hr).
                toLocaleFormat("%X").replace(/:\d\d(?!.*:\d\d)/, "");
        var mlistLocale = document.getElementById("simclk-mlistLocale");
        var location = mlistLocale.label;
        var tree = document.getElementById("simclk-treeNotifications");
        var boxobject = tree.boxObject;
        boxobject.QueryInterface(Components.interfaces.nsITreeBoxObject);
        var treeChildren = document.getElementById("simclk-treeItems");
        var treeItems, row = null;
        var timeNode, localeNode = null;
        var dowValue;
        var descr = document.getElementById("simclk-tboxNotifyDesc").value;
        var recurring = document.getElementById("simclk-cboxNotifyRecurring").checked;
        var updaterow = document.getElementById('simclk-tboxNotifyDesc').
                getAttribute("updaterow");

        if ( updaterow ) {
            tree.disabled = false;

            // Updating a notification, enable all buttons.
            document.getElementById("simclk-btnDel").disabled = false;
            document.getElementById("simclk-btnDelAll").disabled = false;
            document.getElementById("simclk-btnReset").disabled = false;

            // Reset label and tooltip on Add and Add/OK buttons.
            document.getElementById("simclk-btnAddOk").label =
                    strbundle.getString("button.label.add/OK");
            document.getElementById("simclk-btnAddOk").tooltipText =
                        strbundle.getString("button.tooltip.add/OK");
            document.getElementById("simclk-btnAdd").label =
                    strbundle.getString("button.label.add");
            document.getElementById("simclk-btnAdd").tooltipText =
                        strbundle.getString("button.tooltip.add");
        }

        // If URL present, check for valid protocol.
        var url = document.getElementById("simclk-tboxNotifyUrl").value;

        if ( url ) {
            // Strip whitespace.
            newUrl = url.replace(/[\n\t ]+/, "");
            if ( this.validateUrl(newUrl) ) {
                document.getElementById("simclk-tboxNotifyUrl").value = newUrl;
                url = newUrl;
            }
            else {
                alert(strbundle.getString("alert.error.invalid.url"));
                return false;
            }
        }

        if ( recurring ) {
            // Default dow is daily.
            var mlistDow = document.getElementById("simclk-mlistDow");
            var dowSelected = mlistDow.selectedItem;
            dowValue = dowSelected ?
                    dowSelected.value : "7";  // daily = 7
        }
        else {
            dowValue = "-1";
        }

        // Check if the new/updated notification time is a duplicate.
        if ( treeChildren.hasChildNodes() ) {
            var dup = false;
            treeItems = treeChildren.childNodes;
            tree.view.selection.clearSelection();
            var len = treeItems.length;

            for ( var i = 0; i < len; i++ ) {
                row = treeItems[i].firstChild;
                timeNode = row.firstChild;
                localeNode = timeNode.nextSibling;
                dayNode = localeNode.nextSibling.nextSibling;

                if ( timeDisplay ===  timeNode.getAttribute("label") &&
                     location === localeNode.getAttribute("label") &&
                     parseInt(updaterow, 10) !== i ) {
                    // Time/location match.
                    if ( dowValue === dayNode.getAttribute("value") ||
                         this.checkDuplicateDay(parseInt(dayNode.getAttribute("value"), 10),
                                                parseInt(dowValue, 10)) ) {
                        // dow match or conflict.
                        // Maybe more than one dup, keep searching and highlight them all.
                        // Parm "true" indicates add to current selection.
                        tree.view.selection.rangedSelect(i, i, true);
                        dup = true;
                    }
                }
            }

            if ( dup ) {
                // Notify user the not. was added, but with conflict(s).
                // Alerts will occur about 3 seconds apart.
                updaterow ? alert(strbundle.getString("alert.warning.duplicate.notification.upd")) :
                            alert(strbundle.getString("alert.warning.duplicate.notification.add"));
            }
        }

        var sortValue = 0;

        if ( mlistLocale.selectedItem ) {
            sortValue = mlistLocale.selectedItem.value;
        }

        var data = {completed: false, recurring: recurring, timeDisplay: timeDisplay,
                description: descr, time24hr: time24hr, timeMilli: 0, timeAdjustedForDst: false,
                currDay: -1, day: dowValue, url: url, location: location, sortValue: sortValue};
        this.addTreeRow(treeChildren, data);

        // Select the added/updated row. Add to any existing selections (duplicates).
        var rowIndex;

        if ( updaterow ) {
            rowIndex = parseInt(updaterow, 10);
        }
        else if ( treeItems ) {
            rowIndex = treeItems.length - 1;
        }

        tree.view.selection.rangedSelect(rowIndex, rowIndex, true);

        // Ensure the added/updated row is visible, tree displays 5 rows.
        if ( treeItems && treeItems.length > 5 ) {
            boxobject.ensureRowIsVisible(rowIndex);
        }

        // Clear the textboxes.
        document.getElementById("simclk-tboxNotifyDesc").value = "";
        document.getElementById("simclk-tboxNotifyUrl").value = "";

        return true;
    },

    // Add a notification to the tree.

    addTreeRow: function(treeChildren, data) {
        var item = document.createElement("treeitem");
        var row = document.createElement("treerow");
        var cell;

        // First cell is "timeDisplay", not editable. Stash "time24hr" in value.
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.timeDisplay);
        cell.setAttribute("value", data.time24hr);
        cell.setAttribute("timeMilli", data.timeMilli);
        cell.setAttribute("timeAdjustedForDst", data.timeAdjustedForDst);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Second cell is "locale", not editable.
        // Primary sort on not. table is by location.
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.location);
        cell.setAttribute("value", data.sortValue);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Third cell is "recurring", displayed as a checkbox.
        cell = document.createElement("treecell");
        cell.setAttribute("value", data.recurring);
        row.appendChild(cell);

        // Fourth cell is "day". Stash numeric value of day in value.
        cell = document.createElement("treecell");
        cell.setAttribute("label", SimpleClocksLocations.getDow(parseInt(data.day, 10)));
        cell.setAttribute("value", data.day);
        cell.setAttribute("currDay", data.currDay);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Fifth cell is "URL", the optional URL".
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.url);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Sixth cell is "descr", the optional description.
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.description);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Add the treerow onto treeitem.
        item.appendChild(row);

        // Append or replace? Check our custom attrribute.
        // When called from onLoadNotify, updaterow = "".
        var updaterow = document.getElementById("simclk-tboxNotifyDesc").
                getAttribute("updaterow");

        if ( updaterow ) {
            treeChildren.replaceChild(item, treeChildren.childNodes[parseInt(updaterow, 10)]);
            document.getElementById("simclk-tboxNotifyDesc").setAttribute("updaterow", "");
        }
        else {
            treeChildren.appendChild(item);
        }
    },

    // Called when Delete button is clicked.

    onDeleteNotify: function() {
        var strbundle = document.getElementById("simclk-strings");
        var tree = document.getElementById("simclk-treeNotifications");

        if ( tree.currentIndex < 0 ) {
            // User didn't make a selection.
            alert(strbundle.getString("alert.warning.no.item.selected"));
        }
        else {
            var start = {};
            var end = {};;
            var treeItem;
            var numRanges = tree.view.selection.getRangeCount();

            // Start with the last selection range and work up.
            for ( var i = numRanges - 1; i >= 0; i-- ) {
                tree.view.selection.getRangeAt(i,start,end);

                // Within a range, delete from the bottom up.
                for ( var j = end.value; j >= start.value; j-- ) {
                    treeItem = tree.view.getItemAtIndex(j);
                    treeItem.parentNode.removeChild(treeItem);
                }
            }
        }
    },

    // Called when Del All button is clicked.

    onDeleteAllNotify: function() {
        var treeChildren = document.getElementById("simclk-treeItems");

        // Remove from the bottom.
        while ( treeChildren.hasChildNodes() ) {
            treeChildren.removeChild(treeChildren.lastChild);
        }
    },

    // Check if newly added/updated not. will result in alarms
    // occuring at the same time.

    checkDuplicateDay: function(oldDow, newDow) {
        // Not. day values:
        // 0-6 for Sun-Sat,
        // 7 daily, 8 weekdays, 9 weekends.
        // -1 for non-recurring (ignore them).
        if ( oldDow === 7 || newDow === 7 ||
             (( oldDow === 0 || oldDow === 6 ) && newDow === 9) ||
             (( newDow === 0 || newDow === 6 ) && oldDow === 9) ||
             (( oldDow > 0 && oldDow < 6 ) && newDow === 8) ||
             (( newDow > 0 && newDow < 6 ) && oldDow === 8) ) {
            return true;
        }
        else {
            return false;
        }
    },

    // Called when user adds a notification containing a URL.

    validateUrl: function(url) {
        // Check for supported protocols.
        if ( url.match(/^http:\/\//) ||
             url.match(/^https:\/\//) ||
             url.match(/^file:\/\/\//) ||
             url.match(/^ftp:\/\//) ||
             url.match(/^chrome:\/\//) ) {
            return true;
        }
        else {
            return false;
        }
    },

    // Notify Me At dialog tooltip.
    // Show the time difference between local
    // and other locale. When called from populateLocaleMenulist(),
    // location is passed; when called by event onpopupshowing,
    // event is passed.

    notifyTip: function(event, location) {
        var strbundle = document.getElementById("simclk-strings");
        var locationTime = {time: "", timeDiffHrs: 0};
        var tree = document.getElementById("simclk-treeNotifications");
        var descr = document.getElementById("simclk-descrNotTreeTip");

        if ( event ) {
            // Notificatons tree tooltip.
            // Get the row, col and child element at the point of mouse click.
            var tbo = tree.treeBoxObject;
            var row = { }, col = { }, child = { };
            tbo.getCellAt(event.clientX, event.clientY, row, col, child);

            if ( row.value >= 0) {
                location = tree.view.getCellText(row.value,
                    tree.columns.getNamedColumn("simclk-colLocale"));

                if ( location === strbundle.getString("msg.local") ) {
                    // Don't show tooltip if local.
                    return false;
                }
            }
            else {
                return false;
            }
        }

        var locationFound = false;

        for ( var i in this.locationsTable ) {
            if ( this.locationsTable[i].location === location ) {
                locationFound = true;
                break;
            }
        }

        if ( !locationFound ) {
            return false;
        }

        locationTime = SimpleClocksLocations.getLocationTime(this.locationsTable[i], true);

        if ( locationTime.timeDiffHrs > 0 ) {
            descr.value = location + " " + strbundle.getString("msg.time.ahead").
                    replace("%h", locationTime.timeDiffHrs);
        }
        else if ( locationTime.timeDiffHrs < 0 ) {
            descr.value = location + " " + strbundle.getString("msg.time.behind").
                    replace("%h", Math.abs(locationTime.timeDiffHrs));
        }
        else {
            descr.value = location + " " + strbundle.getString("msg.time.same");
        }

        return descr.value;
    },

    // Called when dialog loaded.
    // Populate the Locale menulist popup.

    populateLocaleMenulist: function() {
        var sortValue, tempItem;

        if ( this.locationsTable.length > 0 ) {
            var localePopup = document.getElementById("simclk-mpopLocale");
            this.clearMenu(localePopup);
        }

        for ( var i in this.locationsTable ) {
            // Value used for sorting.
            sortValue = parseInt(i, 10) + 1;

            tempItem = document.createElement("menuitem");
            tempItem.setAttribute("label", this.locationsTable[i].location);
            tempItem.setAttribute("value", sortValue);
            var ttText = this.notifyTip("", this.locationsTable[i].location);

            if ( ttText ) {
                tempItem.setAttribute("tooltiptext", ttText);
            }

            localePopup.appendChild(tempItem);
            this.refreshNotifySortValues(this.locationsTable[i].location, sortValue);
        }
    },

    // Remove existing menulist popup items, but leave root item.

    clearMenu: function (popup) {
        for ( var i = popup.childNodes.length - 1; i > 0; i-- ) {
             popup.removeChild(popup.childNodes.item(i));
        }
    },

    // Refresh the sort value to match current locale menulist.

    refreshNotifySortValues: function (location, sortValue) {
        var strbundle = document.getElementById("simclk-strings");

        for ( var i in this.notifyTable ) {
            if ( !this.notifyTable[i].completed &&
                  this.notifyTable[i].location !== strbundle.getString("msg.local") ) {
                if ( this.notifyTable[i].location === location ) {
                    this.notifyTable[i].sortValue = sortValue;
                }
            }
        }
    },

    // Called when user clicks OK.

    onOKNotify: function() {
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                getService(Components.interfaces.nsIPrefService).
                getBranch("extensions.simpletimerClocks@grbradt.org.");

        this.buildNotifyTable();
        this.updateNotifyPref();

        // Trigger pref observer, which starts not. processing.
        // Not done in updateNotifyPref() since its also called from SimpleClocks.
        prefs.setBoolPref("notifyClosed", !prefs.getBoolPref("notifyClosed"));
    },

    // Construct an array of notification objects.
    // Called when the user clicks OK or Add/OK.

    buildNotifyTable: function() {
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        var treeChildren = document.getElementById("simclk-treeItems");
        var notifyTable = [];

        if ( treeChildren.hasChildNodes() ) {
            // Build table of all nots. from dialog not. tree.
            var row, time24hr, completed, recurring, description,
                day, url, timeNode, timeDisplay, timeMilli, timeAdjustedForDst,
                currDay, dayNode, locationNode, location, sortValue;

            var treeItems = treeChildren.childNodes;
            var len = treeItems.length;

            for ( var i = 0; i < len; i++ ) {
                row = treeItems[i].firstChild;
                timeNode = row.firstChild;
                timeDisplay = timeNode.getAttribute("label");
                timeMilli = parseInt(timeNode.getAttribute("timeMilli"), 10);
                timeAdjustedForDst = ( timeNode.getAttribute("timeAdjustedForDst") === "true" ) ?
                        true : false;

                // Zero pad the hours if necessary.
                time24hr = timeNode.getAttribute("value");
                time24hr = ( time24hr.indexOf(":") === 1 ) ?
                        "0" + time24hr : time24hr;

                locationNode = row.firstChild.nextSibling;
                location = locationNode.getAttribute("label");
                sortValue = parseInt(locationNode.getAttribute("value"), 10);

                // Convert string attribute to boolean.
                recurring = ( row.firstChild.nextSibling.nextSibling.
                                getAttribute("value") === "true" ) ?
                        true : false;

                dayNode = row.lastChild.previousSibling.previousSibling;
                day = parseInt(dayNode.getAttribute("value"), 10);
                currDay = parseInt(dayNode.getAttribute("currDay"), 10);

                url = row.lastChild.previousSibling.getAttribute("label");

                description = row.lastChild.getAttribute("label");

                notifyTable[i] = {completed: false,
                                  recurring: recurring,
                                  timeDisplay: timeDisplay,
                                  description: description,
                                  time24hr: time24hr,
                                  timeMilli: timeMilli,
                                  timeAdjustedForDst: timeAdjustedForDst,
                                  currDay: currDay,
                                  day: day,
                                  url: url,
                                  location: location,
                                  sortValue: sortValue
                                 };
            }
        }

        // Sort the table on location, time24hr properties.
        if ( notifyTable.length > 1 ) {
            notifyTable.sort(this.sortNotifyTableLocation);
            notifyTable.sort(this.sortNotifyTableTime);
        }

        Application.storage.set("notifyTable", notifyTable);
    },

    // Construct an array of notification objects from the notifications pref,
    // and store via FUEL. Called at startup.

    buildNotifyTableFromPref: function(notifications) {
        var strbundle = document.getElementById("simclk-strings");

        // Objects stored as JSON strings in the pref.
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);

        var notExpired = false;
        var description = "";
        var expiredArray = [];
        var notifyTable = [];
        var notifyArray = notifications.split(this.notifySep);

        for ( var i in notifyArray ) {
            notifyTable[i] = json.decode(notifyArray[i]);

            // This code from Tsprajna.
            // All local notifications, of which sortValue == 0,
            // need to change their location to "msg.local".
            // So that switching Locale won't affect these notifications.
            if ( notifyTable[i].sortValue === 0 &&
                 notifyTable[i].location !== strbundle.getString("msg.local") ) {
                notifyTable[i].location = strbundle.getString("msg.local");
            }

            // Check if a non-recurring not. expired while user
            // was logged off, if so mark completed.
            if ( !notifyTable[i].recurring &&
                 !notifyTable[i].completed &&
                  new Date().getTime() > notifyTable[i].timeMilli ) {
                notExpired = true;
                notifyTable[i].completed = true;
                description = (notifyTable[i].description) ?
                        notifyTable[i].location + " / " + notifyTable[i].description :
                        notifyTable[i].location;

                expiredArray.push({timeDisplay: notifyTable[i].timeDisplay,
								   description: description});

                if ( SimpleClocks.eventLogging ) {
                    var url = ( notifyTable[i].url ) ?
                            notifyTable[i].url :
                            strbundle.getString("msg.none");

					// Pass event type, event time, recurring, status, description, URL.
					SimpleClocksEventLog.logEvent(strbundle.getString("msg.notification"),
												  notifyTable[i].timeDisplay,
                                                  strbundle.getString("msg.no"),
                                                  strbundle.getString("msg.expired"),
												  description,
												  url);
				}
            }
        }

        // Sort the table on location, time24hr properties.
        if ( notifyTable.length > 1 ) {
            notifyTable.sort(this.sortNotifyTableLocation);
            notifyTable.sort(this.sortNotifyTableTime);
        }

        Application.storage.set("notifyTable", notifyTable);

        if ( notExpired ) {
            this.updateNotifyPref();

            if ( !SimpleClocksSliderAlert.otherBrowserWindowOpen() ) {
                var params = { displayItems: expiredArray, msg: strbundle.getString("msg.notify.lost")};
                setTimeout(SimpleClocksSliderAlert.addMessageToQueue, 5000, params);
                setTimeout("SimpleClocks.playSound(3)", 5000);
            }
        }
    },

    // Write notifications pref from not. table. Also called from
    // SimpleClocks.setNotifyInterval().

    updateNotifyPref: function() {
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                getService(Components.interfaces.nsIPrefService).
                getBranch("extensions.simpletimerClocks@grbradt.org.");

        // For Unicode.
        var str = Components.classes["@mozilla.org/supports-string;1"].
                createInstance(Components.interfaces.nsISupportsString);

        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        var notifyTable = [];

        if ( Application.storage.has("notifyTable") ) {
            notifyTable = Application.storage.get("notifyTable", "ERR");

            if ( notifyTable === "ERR" ) {
                alert(strbundle.getString("alert.error.loading.notifications"));
                return;
            }
        }
        else {
            alert(strbundle.getString("alert.error.loading.notifications"));
            return;
        }

        // Objects stored as JSON strings in the pref.
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);

        var prefString = "";

        for ( var i in notifyTable ) {
            prefString += json.encode(notifyTable[i]) + this.notifySep;
        }

        // Remove last separator.
        if ( prefString ) {
            prefString = prefString.substring(0, prefString.length - 1);
        }

        str.data = prefString;
        prefs.setComplexValue("notifications",
                Components.interfaces.nsISupportsString, str);
    },

    // Part of the tz data update processing.
    // Resetting of notification times will force them to be recalculated
    // in calcNotTimeMilli, using potentially new location data.

    resetNotifyPref: function(notifications) {
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                getService(Components.interfaces.nsIPrefService).
                getBranch("extensions.simpletimerClocks@grbradt.org.");

        // For Unicode.
        var str = Components.classes["@mozilla.org/supports-string;1"].
                createInstance(Components.interfaces.nsISupportsString);

        // Objects stored as JSON strings in the pref.
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);

        var prefString = "";
        var notifyTable = [];
        var notifyArray = notifications.split(this.notifySep);

        for ( var i in notifyArray ) {
            notifyTable[i] = json.decode(notifyArray[i]);
            notifyTable[i].timeMilli = 0;
            notifyTable[i].timeAdjustedForDst = false;
            notifyTable[i].currDay = -1;
            prefString += json.encode(notifyTable[i]) + this.notifySep;
        }

        // Remove last separator.
        if ( prefString ) {
            prefString = prefString.substring(0, prefString.length - 1);
        }

        str.data = prefString;
        prefs.setComplexValue("notifications",
                Components.interfaces.nsISupportsString, str);
    },

    // Sort the notification table on objects sortValue property.
    // This property is bound to the location property.

    sortNotifyTableLocation: function(a, b) {
        return ( a.sortValue - b.sortValue );
    },

    // Sort the notification table on objects time24hr property.

    sortNotifyTableTime: function(a, b) {
        if ( a.sortValue === b.sortValue ) {
            var x = a.time24hr;
            var y = b.time24hr;

            return ( (x < y) ? -1 : ( (x > y) ? 1 : 0 ) );
        }
        else {
            return 0;
        }
    },

    openLink: function(aPage) {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].
                getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var win = wmed.getMostRecentWindow("navigator:browser");

        if ( !win ) {
            win = window.openDialog(
                    "chrome://browser/content/browser.xul",
                    "_blank",
                    "chrome, all, dialog=no",
                    aPage, null, null);
        }
        else {
            var content = win.document.getElementById("content");
            content.selectedTab = content.addTab(aPage);
        }
    },

    // Dump to console.

    dumpNotificationObject: function(notification) {
        this.debug("\nNotification data:");
        this.debug("Completed?: " + notification.completed);
        this.debug("Recurring?: " + notification.recurring);
        this.debug("Notification time: " + notification.timeDisplay);
        this.debug("Description: " + notification.description);
        this.debug("Notification time (24 hr format): " + notification.time24hr);
        this.debug("Notification time (millisecs): " + notification.timeMilli);
        this.debug("Notification time adjusted for dst?: " + notification.timeAdjustedForDst);
        this.debug("Notification day for current timeMilli: " + notification.currDay);
        this.debug("Notification day: " + notification.day);
        this.debug("Url: " + notification.url);
        this.debug("Locale: " + notification.location);
        this.debug("Sort value: " + notification.sortValue);
    },

    // Debug messages to console.

    debug: function(aMsg) {
        setTimeout(function() { throw new Error("[debug] " + aMsg); }, 0);
    }
};
